from django.contrib.auth import get_user_model
from django.contrib.auth.backends import ModelBackend
from django.core.exceptions import FieldDoesNotExist

UserModel = get_user_model()


class SAMLAuthenticationBackend(ModelBackend):
    def authenticate(self, request, idp=None, saml=None):
        nameid = idp.get_nameid(saml)
        attrs = idp.mapped_attributes(saml)
        user, created = UserModel._default_manager.get_or_create(**{UserModel.USERNAME_FIELD: nameid})
        update_fields = []
        for field, values in attrs.items():
            try:
                f = UserModel._meta.get_field(field)
                setattr(user, f.name, values[0])
                update_fields.append(f.name)
            except FieldDoesNotExist:
                pass
        if update_fields:
            user.save(update_fields=update_fields)
        return user



from __future__ import unicode_literals

from django.contrib import admin

from .models import IdP, IdPAttribute


class IdPAttributeInline(admin.TabularInline):
    model = IdPAttribute
    extra = 0


class IdPAdmin(admin.ModelAdmin):
    list_display = (
        "name",
        "slug",
        "last_import",
        "certificate_expires",
        "get_entity_id",
        "get_acs",
        "is_active",
        "last_login",
    )
    list_filter = ("is_active",)
    actions = ("import_metadata", "generate_certificates")
    inlines = (IdPAttributeInline,)
    fieldsets = (
        (None, {"fields": ("name", "slug", "base_url", "notes", "is_active")}),
        (
            "SP Settings",
            {"fields": ("contact_name", "contact_email", "x509_certificate", "private_key", "certificate_expires")},
        ),
        (
            "IdP Metadata",
            {"fields": ("metadata_url", "verify_metadata_cert", "metadata_xml", "lowercase_encoding", "last_import")},
        ),
        ("Logins", {"fields": ("respect_expiration", "login_redirect", "last_login")}),
        ("Advanced", {"classes": ("collapse",), "fields": ("authenticate_method", "login_method")}),
    )
    prepopulated_fields = {"slug": ("name",)}
    readonly_fields = ("last_import", "last_login")

    def get_changeform_initial_data(self, request):
        return {"base_url": "{}://{}".format(request.scheme, request.get_host())}

    def generate_certificates(self, request, queryset):
        for idp in queryset:
            idp.generate_certificate()

    def import_metadata(self, request, queryset):
        for idp in queryset:
            idp.import_metadata()

    def save_model(self, request, obj, form, change):
        super(IdPAdmin, self).save_model(request, obj, form, change)
        try:
            obj.import_metadata()
        except Exception:
            pass


admin.site.register(IdP, IdPAdmin)



from django.urls import path

from . import views

urlpatterns = [
    path("<idp_slug>/", views.metadata, name="sp-idp-metadata"),
    path("<idp_slug>/acs/", views.acs, name="sp-idp-acs"),
    path("<idp_slug>/login/", views.login, name="sp-idp-login"),
    path("<idp_slug>/test/", views.login, {"test": True}, name="sp-idp-test"),
    path("<idp_slug>/verify/", views.login, {"verify": True}, name="sp-idp-verify"),
]



# Generated by Django 2.2.7 on 2019-11-13 15:49

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = []

    operations = [
        migrations.CreateModel(
            name="IdP",
            fields=[
                ("id", models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                ("name", models.CharField(max_length=200)),
                ("slug", models.CharField(max_length=100, unique=True)),
                (
                    "base_url",
                    models.CharField(
                        help_text="Root URL for the site, including http/https, no trailing slash.", max_length=200
                    ),
                ),
                ("contact_name", models.CharField(max_length=100)),
                ("contact_email", models.EmailField(max_length=100)),
                ("x509_certificate", models.TextField(blank=True)),
                ("private_key", models.TextField(blank=True)),
                ("certificate_expires", models.DateTimeField(blank=True, null=True)),
                (
                    "metadata_url",
                    models.URLField(
                        blank=True,
                        help_text="Leave this blank if entering metadata XML directly.",
                        max_length=500,
                        verbose_name="Metadata URL",
                    ),
                ),
                (
                    "verify_metadata_cert",
                    models.BooleanField(default=True, verbose_name="Verify metadata URL certificate"),
                ),
                (
                    "metadata_xml",
                    models.TextField(
                        blank=True,
                        help_text="Automatically loaded from the metadata URL, if specified. Otherwise input directly.",
                        verbose_name="Metadata XML",
                    ),
                ),
                (
                    "lowercase_encoding",
                    models.BooleanField(default=False, help_text="Check this if the identity provider is ADFS."),
                ),
                (
                    "saml_settings",
                    models.TextField(
                        blank=True, editable=False, help_text="Settings imported and used by the python-saml library."
                    ),
                ),
                ("last_import", models.DateTimeField(blank=True, null=True)),
                ("notes", models.TextField(blank=True)),
                (
                    "respect_expiration",
                    models.BooleanField(
                        default=False,
                        help_text="Expires the Django session based on the IdP session expiration. Only works when using SESSION_SERIALIZER=PickleSerializer.",
                        verbose_name="Respect IdP session expiration",
                    ),
                ),
                (
                    "login_redirect",
                    models.CharField(
                        blank=True, help_text="URL name or path to redirect after a successful login.", max_length=200
                    ),
                ),
                ("last_login", models.DateTimeField(blank=True, default=None, null=True)),
                ("is_active", models.BooleanField(default=True)),
                ("authenticate_method", models.CharField(default="sp.utils.authenticate", max_length=200)),
                ("login_method", models.CharField(default="sp.utils.login", max_length=200)),
            ],
            options={"verbose_name": "identity provider",},
        ),
        migrations.CreateModel(
            name="IdPAttribute",
            fields=[
                ("id", models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                ("saml_attribute", models.CharField(max_length=200)),
                ("mapped_name", models.CharField(blank=True, max_length=200)),
                ("is_nameid", models.BooleanField(default=False, verbose_name="Is NameID")),
                (
                    "idp",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="attributes",
                        to="sp.IdP",
                        verbose_name="identity provider",
                    ),
                ),
            ],
            options={
                "verbose_name": "identity provider attribute",
                "verbose_name_plural": "identity provider attributes",
                "unique_together": {("idp", "saml_attribute")},
            },
        ),
    ]



import datetime

from django.conf import settings
from django.contrib import auth
from django.core.exceptions import ImproperlyConfigured

from .models import IdP

IDP_SESSION_KEY = "_idp_slug"


def authenticate(request, idp, saml):
    return auth.authenticate(request, idp=idp, saml=saml)


def login(request, user, idp, saml):
    auth.login(request, user)
    request.session[IDP_SESSION_KEY] = idp.slug
    if idp.respect_expiration:
        if not settings.SESSION_SERIALIZER.endswith("PickleSerializer"):
            raise ImproperlyConfigured(
                "IdP-based session expiration is only supported with the PickleSerializer SESSION_SERIALIZER."
            )
        try:
            dt = datetime.datetime.fromtimestamp(saml.get_session_expiration(), tz=datetime.timezone.utc)
            request.session.set_expiry(dt)
        except TypeError:
            pass


def get_session_idp(request):
    return IdP.objects.filter(slug=request.session.get(IDP_SESSION_KEY)).first()



from onelogin.saml2.auth import OneLogin_Saml2_Auth
from onelogin.saml2.settings import OneLogin_Saml2_Settings

from django.contrib.auth import REDIRECT_FIELD_NAME
from django.core import signing
from django.http import HttpResponse
from django.shortcuts import get_object_or_404, redirect, render
from django.utils import timezone
from django.views.decorators.csrf import csrf_exempt
from django.views.decorators.http import require_POST

from .models import IdP


def metadata(request, idp_slug):
    idp = get_object_or_404(IdP, slug=idp_slug, is_active=True)
    saml_settings = OneLogin_Saml2_Settings(settings=idp.sp_settings, sp_validation_only=True)
    return HttpResponse(saml_settings.get_sp_metadata(), content_type="text/xml")


@csrf_exempt
@require_POST
def acs(request, idp_slug):
    idp = get_object_or_404(IdP, slug=idp_slug, is_active=True)
    if request.POST.get("RelayState"):
        # Login with state relayed from our application.
        state = signing.loads(request.POST["RelayState"], max_age=300)
    else:
        # IdP-initiated login.
        state = {"test": False, "verify": False, "redir": ""}
    saml = OneLogin_Saml2_Auth(idp.prepare_request(request), old_settings=idp.settings)
    saml.process_response()
    errors = saml.get_errors()
    if errors:
        return render(
            request,
            "sp/error.html",
            {"idp": idp, "state": state, "errors": errors, "reason": saml.get_last_error_reason()},
            status=500,
        )
    else:
        if state.get("test", False):
            attrs = []
            for saml_attr, value in saml.get_attributes().items():
                attr, created = idp.attributes.get_or_create(saml_attribute=saml_attr)
                attrs.append((attr, "; ".join(value)))
            return render(
                request,
                "sp/test.html",
                {"idp": idp, "attrs": attrs, "nameid": saml.get_nameid(), "nameid_format": saml.get_nameid_format()},
            )
        elif state.get("verify", False):
            user = idp.authenticate(request, saml)
            if user == request.user:
                return redirect(idp.get_login_redirect(state.get("redir")))
            else:
                return render(request, "sp/unauth.html", {"idp": idp, "verify": True}, status=401)
        else:
            user = idp.authenticate(request, saml)
            if user:
                idp.login(request, user, saml)
                idp.last_login = timezone.now()
                idp.save(update_fields=("last_login",))
                return redirect(idp.get_login_redirect(state.get("redir")))
            else:
                return render(request, "sp/unauth.html", {"idp": idp, "verify": False}, status=401)


def login(request, idp_slug, test=False, verify=False):
    idp = get_object_or_404(IdP, slug=idp_slug, is_active=True)
    saml = OneLogin_Saml2_Auth(idp.prepare_request(request), old_settings=idp.settings)
    reauth = verify or "reauth" in request.GET
    state = signing.dumps({"test": test, "verify": verify, "redir": request.GET.get(REDIRECT_FIELD_NAME, "")})
    return redirect(saml.login(state, force_authn=reauth))



import collections
import datetime
import json
from urllib.parse import urlparse

from cryptography import x509
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import hashes, serialization
from cryptography.hazmat.primitives.asymmetric import rsa
from cryptography.x509.oid import NameOID
from onelogin.saml2.idp_metadata_parser import OneLogin_Saml2_IdPMetadataParser

from django.conf import settings
from django.db import models
from django.urls import reverse
from django.utils import timezone
from django.utils.module_loading import import_string
from django.utils.translation import ugettext_lazy as _


class IdP(models.Model):
    name = models.CharField(max_length=200)
    slug = models.CharField(max_length=100, unique=True)
    base_url = models.CharField(
        max_length=200, help_text=_("Root URL for the site, including http/https, no trailing slash.")
    )
    contact_name = models.CharField(max_length=100)
    contact_email = models.EmailField(max_length=100)
    x509_certificate = models.TextField(blank=True)
    private_key = models.TextField(blank=True)
    certificate_expires = models.DateTimeField(null=True, blank=True)
    metadata_url = models.URLField(
        "Metadata URL", max_length=500, blank=True, help_text=_("Leave this blank if entering metadata XML directly.")
    )
    verify_metadata_cert = models.BooleanField(_("Verify metadata URL certificate"), default=True)
    metadata_xml = models.TextField(
        _("Metadata XML"),
        blank=True,
        help_text=_("Automatically loaded from the metadata URL, if specified. Otherwise input directly."),
    )
    lowercase_encoding = models.BooleanField(default=False, help_text=_("Check this if the identity provider is ADFS."))
    saml_settings = models.TextField(
        blank=True, help_text=_("Settings imported and used by the python-saml library."), editable=False
    )
    last_import = models.DateTimeField(null=True, blank=True)
    notes = models.TextField(blank=True)
    respect_expiration = models.BooleanField(
        _("Respect IdP session expiration"),
        default=False,
        help_text=_(
            "Expires the Django session based on the IdP session expiration. Only works when using SESSION_SERIALIZER=PickleSerializer."
        ),
    )
    login_redirect = models.CharField(
        max_length=200, blank=True, help_text=_("URL name or path to redirect after a successful login.")
    )
    last_login = models.DateTimeField(null=True, blank=True, default=None)
    is_active = models.BooleanField(default=True)
    authenticate_method = models.CharField(max_length=200, default="sp.utils.authenticate")
    login_method = models.CharField(max_length=200, default="sp.utils.login")

    class Meta:
        verbose_name = _("identity provider")

    def __str__(self):
        return self.name

    def get_entity_id(self):
        return self.base_url + self.get_absolute_url()

    get_entity_id.short_description = _("Entity ID")

    def get_acs(self):
        return self.base_url + reverse("sp-idp-acs", kwargs={"idp_slug": self.slug})

    get_acs.short_description = _("ACS")

    def get_absolute_url(self):
        return reverse("sp-idp-metadata", kwargs={"idp_slug": self.slug})

    def get_login_url(self):
        return reverse("sp-idp-login", kwargs={"idp_slug": self.slug})

    def get_test_url(self):
        return reverse("sp-idp-test", kwargs={"idp_slug": self.slug})

    def get_verify_url(self):
        return reverse("sp-idp-verify", kwargs={"idp_slug": self.slug})

    def prepare_request(self, request):
        return {
            "https": "on" if request.is_secure() else "off",
            "http_host": request.get_host(),
            "script_name": request.path_info,
            "server_port": 443 if request.is_secure() else request.get_port(),
            "get_data": request.GET.copy(),
            "post_data": request.POST.copy(),
            "lowercase_urlencoding": self.lowercase_encoding,
        }

    @property
    def sp_settings(self):
        return {
            "strict": True,
            "sp": {
                "NameIDFormat": "urn:oasis:names:tc:SAML:1.1:nameid-format:unspecified",
                "entityId": self.get_entity_id(),
                "assertionConsumerService": {
                    "url": self.get_acs(),
                    "binding": "urn:oasis:names:tc:SAML:2.0:bindings:HTTP-POST",
                },
                "x509cert": self.x509_certificate,
                "privateKey": self.private_key,
            },
            "contactPerson": {"technical": {"givenName": self.contact_name, "emailAddress": self.contact_email}},
        }

    @property
    def settings(self):
        settings_dict = json.loads(self.saml_settings)
        settings_dict.update(self.sp_settings)
        return settings_dict

    def generate_certificate(self):
        url_parts = urlparse(self.base_url)
        backend = default_backend()
        key = rsa.generate_private_key(public_exponent=65537, key_size=2048, backend=backend)
        self.private_key = key.private_bytes(
            encoding=serialization.Encoding.PEM,
            format=serialization.PrivateFormat.PKCS8,
            encryption_algorithm=serialization.NoEncryption(),
        ).decode("ascii")
        name = x509.Name([x509.NameAttribute(NameOID.COMMON_NAME, url_parts.netloc)])
        basic_contraints = x509.BasicConstraints(ca=True, path_length=0)
        now = timezone.now()
        self.certificate_expires = now + datetime.timedelta(days=3650)
        cert = (
            x509.CertificateBuilder()
            .subject_name(name)
            .issuer_name(name)
            .public_key(key.public_key())
            .serial_number(x509.random_serial_number())
            .not_valid_before(now)
            .not_valid_after(self.certificate_expires)
            .add_extension(basic_contraints, critical=False)
            .sign(key, hashes.SHA1(), backend)
        )
        self.x509_certificate = cert.public_bytes(serialization.Encoding.PEM).decode("ascii")
        self.save()

    def import_metadata(self):
        if self.metadata_url:
            self.metadata_xml = OneLogin_Saml2_IdPMetadataParser.get_metadata(
                self.metadata_url, validate_cert=self.verify_metadata_cert
            ).decode("utf-8")
        self.saml_settings = json.dumps(OneLogin_Saml2_IdPMetadataParser.parse(self.metadata_xml))
        self.last_import = timezone.now()
        self.save()

    def mapped_attributes(self, saml):
        attrs = collections.OrderedDict()
        for attr in self.attributes.exclude(mapped_name=""):
            value = saml.get_attribute(attr.saml_attribute)
            if value is not None:
                attrs[attr.mapped_name] = value
        return attrs

    def get_nameid(self, saml):
        nameid_attr = self.attributes.filter(is_nameid=True).first()
        if nameid_attr:
            return saml.get_attribute(nameid_attr.saml_attribute)[0]
        else:
            return saml.get_nameid()

    def get_login_redirect(self, redir=None):
        return redir or self.login_redirect or settings.LOGIN_REDIRECT_URL

    def authenticate(self, request, saml):
        return import_string(self.authenticate_method)(request, self, saml)

    def login(self, request, user, saml):
        return import_string(self.login_method)(request, user, self, saml)


class IdPAttribute(models.Model):
    idp = models.ForeignKey(
        IdP, verbose_name=_("identity provider"), related_name="attributes", on_delete=models.CASCADE
    )
    saml_attribute = models.CharField(max_length=200)
    mapped_name = models.CharField(max_length=200, blank=True)
    is_nameid = models.BooleanField(_("Is NameID"), default=False)

    class Meta:
        verbose_name = _("identity provider attribute")
        verbose_name_plural = _("identity provider attributes")
        unique_together = [
            ("idp", "saml_attribute"),
        ]

    def __str__(self):
        if self.mapped_name:
            return "{} -> {}".format(self.saml_attribute, self.mapped_name)
        else:
            return "{} (unmapped)".format(self.saml_attribute)



from setuptools import setup

# Note, to build a wheel that does not include testapp (bdist_wheel does not respect find/exclude):
# python setup.py sdist && pip wheel --no-index --no-deps --wheel-dir dist dist/*.tar.gz

setup()
